PLANNER_VERSION = 3.103;
PLANNER_MAX_POINTS = 71;

function Planner_OnLoad(self)
	if ( not IsAddOnLoaded( "Blizzard_TalentUI" ) ) then
		_, reason = LoadAddOn( "Blizzard_TalentUI" );
	end

	if ( reason ) then
		Planner_Log( "Blizzard_TalentUI not loaded. Exiting." );
		return;
	end
	
	-- Hook the Talent Frame's Update function
	hooksecurefunc( "TalentFrame_Update", Planner_TalentFrame_AfterUpdate );
	
	-- Replace the Talent Buttons' OnEnter and OnClick scripts
	for i = 1, MAX_NUM_TALENTS, 1 do
		local button = getglobal( "PlayerTalentFrameTalent"..i );
		button:RegisterForClicks( "LeftButtonUp", "RightButtonUp" );
		
		Planner_TalentFrameTalent_OnClick = button:GetScript( "OnClick" );
		button:SetScript( "OnClick", Planner_OnClick );
		
		Planner_TalentFrameTalent_OnEnter = button:GetScript( "OnEnter" );
		button:SetScript( "OnEnter", Planner_OnEnter );
	end
	
	-- Register for Loading Variables
	self:RegisterEvent( "VARIABLES_LOADED" );
	self:RegisterEvent( "PLAYER_ENTERING_WORLD" );
	self:RegisterEvent( "PLAYER_LEAVING_WORLD" );
	
	SlashCmdList["PLANNER"] = function ( msg )
		Planner_DoSlash( msg );
	end

	SLASH_PLANNER1 = "/planner";
	SLASH_PLANNER2 = "/tp";
	
	-- Default colors
	INVALID     = { r = 1.0, g = 0.1, b = 0.1, };  -- Red
	PLANNED     = { r = 0.1, g = 0.1, b = 1.0, };  -- Blue
	MAXPLANNED  = { r = 1.0, g = 0.1, b = 1.0, };  -- Purple
	BUYMENOW    = { r = 0.1, g = 1.0, b = 0.1, };  -- Green
	MAXEDOUT    = { r = 1.0, g = 0.82, b = 0.0, }; -- Gold
	NORMAL      = { r = 0.8, g = 0.8, b = 0.8, };  -- Grey
	DONTBUYME   = { r = 0.1, g = 1.0, b = 1.0, };  -- Cyan

	-- Locale/Gender-neutral class names FTW
	tp_classes = { [1] = "DEATHKNIGHT",
	               [2] = "DRUID",
	               [3] = "HUNTER",
	               [4] = "MAGE",
	               [5] = "PALADIN",
	               [6] = "PRIEST",
	               [7] = "ROGUE",
	               [8] = "SHAMAN",
	               [9] = "WARLOCK",
	               [10] = "WARRIOR", };
end

function Planner_OnEvent( self, event, ... )
	if ( event == "VARIABLES_LOADED" ) then
		if ( not tp_ver ) then
			tp_ver = 0.0;
		end
		
		if ( not tp_talents ) then
			tp_talents = {};
		end
		
		if ( not tp_saved ) then
			tp_saved = {};
		end
		
		for i = 1, MAX_NUM_TALENTS, 1 do
			local talent = getglobal( "PlayerTalentFrameTalent"..i );
			local rankBorder = getglobal( "PlayerTalentFrameTalent"..i.."RankBorder" );
			rankBorder:SetWidth( 64 );
			
			local goalRankBorder = talent:CreateTexture( "PlannerRankBorder"..i, "OVERLAY" );
			goalRankBorder:SetTexture( rankBorder:GetTexture() );
			goalRankBorder:SetPoint( "CENTER", talent, "BOTTOMLEFT" );
			
			local goalRank = talent:CreateFontString( "PlannerRank"..i, "OVERLAY" );
			goalRank:SetFontObject( GameFontNormalSmall );
			goalRank:SetPoint( "CENTER", talent, "BOTTOMLEFT" );
		end
	elseif ( event == "PLAYER_ENTERING_WORLD" ) then
		tp_name = UnitName( "player" ).." of "..GetRealmName();
		_, class = UnitClass( "player" );
		
		if ( tp_ver ~= PLANNER_VERSION ) then
			Planner_Upgrade();
		end

		for k, v in ipairs( tp_classes ) do
			if ( v == class ) then
				tp_class = k;
			end
		end
		
		if ( not tp_talents[tp_name] ) then
			tp_talents[tp_name] = { { {}, {}, {}, }, { {}, {}, {} } };
		end
	elseif ( tp_name ) then
		if ( event == "PLAYER_LEAVING_WORLD" ) then
			tp_name = nil;
			tp_class = nil;
		end
	end
end

function Planner_DoSlash( msg )
	if ( not msg or msg == "" ) then
		Planner_ShowUsage();
	end
	
	local words = {};
	for word in string.gmatch( msg, "[^ ]+" ) do
		table.insert( words, word );
	end

	local cmd = words[1];
	local prm1 = words[2];
	local prm2 = words[3];
	local prm3 = words[4];
	
	if ( cmd == "help" ) then
		Planner_ShowUsage();
	elseif ( cmd == "standby" ) then
		tp_standby = not tp_standby;
		
		Planner_HideMe( tp_standby );
		
		GameTooltip:Hide();
		TalentFrame_Update( PlayerTalentFrame );
		Planner_Log( PLANNER_PLANNER..( tp_standby and PLANNER_DE or PLANNER_EN ) );
	elseif ( cmd == "reset" ) then
		local spec = PlayerTalentFrame.talentGroup;
		tp_talents[tp_name][spec] = { {}, {}, {}, };
		
		GameTooltip:Hide();
		TalentFrame_Update( PlayerTalentFrame );
		Planner_Log( PLANNER_REPORT_FORNAME..tp_name..PLANNER_REPORT_RESET );
	elseif ( cmd == "current" ) then
		local spec = PlayerTalentFrame.talentGroup;
		tp_talents[tp_name][spec] = { {}, {}, {}, };
		
		for page = 1, 3, 1 do
			for id = 1, MAX_NUM_TALENTS, 1 do
				_, _, _, _, actRank, _, _, _, pvwRank = GetTalentInfo( page, id, nil, nil, spec );
				
				local rank = GetCVarBool("previewTalents") and pvwRank or actRank;
				
				if ( rank > 0 ) then
					tp_talents[tp_name][spec][page][id] = rank;
				end
			end
		end
		
		GameTooltip:Hide();
		TalentFrame_Update( PlayerTalentFrame );
		Planner_Log( PLANNER_REPORT_FORNAME..tp_name..PLANNER_REPORT_CURRENT );
	elseif ( cmd == "save" ) then
		if ( not prm1 or prm1 == "" ) then
			return;
		end
		
		local spec = PlayerTalentFrame.talentGroup;
		
		tp_saved[tp_name.."::"..prm1] = { {}, {}, {}, };
		
		for page = 1, 3, 1 do
			for k, v in pairs( tp_talents[tp_name][spec][page] ) do
				tp_saved[tp_name.."::"..prm1][page][k] = v;
			end
		end
		
		GameTooltip:Hide();
		TalentFrame_Update( PlayerTalentFrame );
		Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_SAVED );
	elseif ( cmd == "load" ) then
		if ( not prm1 or prm1 == "" ) then
			return;
		end
		
		local spec = PlayerTalentFrame.talentGroup;
		
		if ( tp_saved[tp_name.."::"..prm1] ) then
			tp_talents[tp_name][spec] = { {}, {}, {}, };
		
			for page = 1, 3, 1 do
				for k, v in pairs( tp_saved[tp_name.."::"..prm1][page] ) do
					tp_talents[tp_name][spec][page][k] = v;
				end
			end
			
			GameTooltip:Hide();
			TalentFrame_Update( PlayerTalentFrame );
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_LOADED );
		else
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_NOTFOUND );
		end
	elseif ( cmd == "rename" ) then
		if ( not prm1 or prm1 == "" or not prm2 or ptm2 == "" ) then
			return;
		end
		
		if ( tp_saved[tp_name.."::"..prm1] ) then
			tp_saved[tp_name.."::"..prm2] = tp_saved[tp_name.."::"..prm1];
			tp_saved[tp_name.."::"..prm1] = nil;

			GameTooltip:Hide();
			TalentFrame_Update( PlayerTalentFrame );
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_RENAMED.."\""..prm2.."\"" );
		else
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_NOTFOUND );
		end
	elseif ( cmd == "import" ) then
		if ( not prm1 or prm1 == "" or not prm2 or ptm2 == "" ) then
			return;
		end
		
		prm2 = prm2:sub(1,1):upper()..prm2:sub(2):lower();
		
		if ( not prm3 or prm3 == "" ) then
			prm2 = prm2.." of "..GetRealmName();
		else
			prm2 = prm2.." of "..prm3:sub(1,1):upper()..prm3:sub(2):lower();
		end
		
		if ( tp_saved[prm2.."::"..prm1] ) then
			for page = 1, 3, 1 do
				local num = GetNumTalents( page );
				
				for k, v in pairs( tp_saved[prm2.."::"..prm1][page] ) do
					local _, _, _, _, _, maxRank = GetTalentInfo( page, k );
					
					if ( v > maxRank ) then
						Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_WRONGCLASS );
						return;
					end
				end
			end
			
			local spec = PlayerTalentFrame.talentGroup;
			
			tp_talents[tp_name][spec] = { {}, {}, {}, };
			
			for page = 1, 3, 1 do
				for k, v in pairs( tp_saved[prm2.."::"..prm1][page] ) do
					tp_talents[tp_name][spec][page][k] = v;
				end
			end
			
			GameTooltip:Hide();
			TalentFrame_Update( PlayerTalentFrame );
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_IMPORTED..prm2 );
		else
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_ONALT..prm2..PLANNER_REPORT_NOTFOUND );
		end
	elseif ( cmd == "delete" ) then
		if ( not prm1 or prm1 == "" ) then
			return;
		end
		
		if ( tp_saved[tp_name.."::"..prm1] ) then
			tp_saved[tp_name.."::"..prm1] = nil;

			GameTooltip:Hide();
			TalentFrame_Update( PlayerTalentFrame );
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_DELETED );
		else
			Planner_Log( PLANNER_REPORT_SETNAME.."\""..prm1.."\""..PLANNER_REPORT_NOTFOUND );
		end
	elseif ( cmd == "list" ) then
		if ( not prm1 or prm1 == "" ) then
			prm1 = UnitName( "player" );
		else
			prm1 = prm1:sub(1,1):upper()..prm1:sub(2):lower();
		end
		
		if ( not prm2 or prm2 == "" ) then
			prm1 = prm1.." of "..GetRealmName();
		else
			prm1 = prm1.." of "..prm2:sub(1,1):upper()..prm2:sub(2):lower();
		end
		
		Planner_Log( PLANNER_REPORT_LIST..prm1..":" );
		
		local list = { n = 0, };
		local namelen = string.len( prm1 );
		
		for name, set in pairs( tp_saved ) do
			if ( string.sub( name, 1, namelen + 2 ) == prm1.."::" ) then
				table.insert( list, string.sub( name, namelen + 3 ) );
			end
		end
		
		if ( #list == 0 ) then
			Planner_Log( PLANNER_REPORT_NOLIST );
		else
			for k, v in ipairs( list ) do
				Planner_Log( PLANNER_CYAN..v );
			end
		end
		
		list = nil;
	end
end

function Planner_ShowUsage()
	Planner_Log( "Talent Planner" );
	Planner_Log( "Usage: /planner [option] or /tp [option]" );
	Planner_Log( " Options in angle brackets are required, those in square brackets are optional");
	Planner_Log( " All commands act on the current spec only" );
	Planner_Log( "Available options:" );
	Planner_Log( " help - Shows this help text" );
	Planner_Log( " standby - Enables/Disables the mod" );
	Planner_Log( " reset - Clears your talent goals" );
	Planner_Log( " current - Sets your talent goals to those talents you currently have" );
	Planner_Log( " save <name> - Saves your current goals under the name <name>" );
	Planner_Log( " load <name> - Loads the goals saved as <name>" );
	Planner_Log( " rename <old> <new> - Renames the goals saved as <old> to <new>" );
	Planner_Log( " import <name> <alt> [realm] - Imports the goals saved as <name> from the alt named <alt> on realm [realm]" );
	Planner_Log( " delete <name> - Deletes the goals saved as <name>" );
	Planner_Log( " list [alt] [realm] - Lists saved goals for the alt named [alt] on realm [realm] or for yourself if no alt specified" );
end

function Planner_HideMe( hide )
	for i = 1, MAX_NUM_TALENTS, 1 do
		local rankBorder = getglobal( "PlayerTalentFrameTalent"..i.."RankBorder" );
		local goalRank = getglobal( "PlannerRank"..i );
		local goalRankBorder = getglobal( "PlannerRankBorder"..i );

		if ( hide ) then
			rankBorder:SetWidth( 32 );
			goalRank:Hide();
			goalRankBorder:Hide();
		else
			rankBorder:SetWidth( 64 );
			goalRank:Show();
			goalRankBorder:Show();
		end
	end
end

function Planner_GetCount( page, row )
	local count = 0;
	local spec = PlayerTalentFrame.talentGroup;
	
	if ( page and row ) then
		for i = 1, MAX_NUM_TALENTS, 1 do
			_, _, tier = GetTalentInfo( page, i );
			if ( tier < row and tp_talents[tp_name][spec][page][i] ) then
				count = count + tp_talents[tp_name][spec][page][i];
			end
		end
	elseif ( page ) then
		return Planner_GetCount( page, MAX_NUM_TALENT_TIERS );
	else
		return Planner_GetCount( 1 ) + Planner_GetCount( 2 ) + Planner_GetCount( 3 );
	end
	
	return count;
end

function Planner_GetRank( page, id )
	local spec = PlayerTalentFrame.talentGroup;
	
	if ( tp_talents[tp_name][spec][page] ) then
		return tp_talents[tp_name][spec][page][id] or 0;
	end
end

function Planner_GetPrereq( page, id )
	local reqRow, reqCol = GetTalentPrereqs( page, id );
	
	if ( reqRow and reqCol ) then
		reqIndex = PlayerTalentFrame.TALENT_BRANCH_ARRAY[reqRow][reqCol].id;
		reqName, _, _, _, _, reqNum = GetTalentInfo( page, reqIndex );

		return reqIndex, reqName, reqNum;
	end
end

function Planner_TalentFrame_AfterUpdate( frame )
	local page = PanelTemplates_GetSelectedTab( PlayerTalentFrame );
	local numTalents = GetNumTalents( page );
	local inspect = PlayerTalentFrame.inspect;
	local pet = PlayerTalentFrame.pet;
	local spec = PlayerTalentFrame.talentGroup;
	
	if ( pet ) then
		Planner_HideMe( true );
		return;
	end

	if ( tp_standby or not tp_name or not tp_class or frame ~= PlayerTalentFrame or inspect ) then
		return;
	end

	for id = 1, MAX_NUM_TALENTS, 1 do
		if ( id <= numTalents ) then
			local _, _, _, _, actRank, maxRank, _, _, pvwRank = GetTalentInfo( page, id, nil, nil, spec );
			local rank = GetCVarBool("previewTalents") and pvwRank or actRank;
			local tabName, _, actPoints, _, pvwPoints = GetTalentTabInfo( page, nil, nil, spec );
			local points = actPoints + pvwPoints;
			local freePoints = GetUnspentTalentPoints(nil, nil, spec) - GetGroupPreviewTalentPointsSpent(nil, spec);
			local goalRank = Planner_GetRank( page, id );
			local pagePlanned = Planner_GetCount( page );
			local totalPlanned = Planner_GetCount();
			local isGlowing = getglobal( "PlayerTalentFrameTalent"..id.."RankBorder" ):IsVisible();
			local color = nil;

			if ( rank > goalRank or goalRank > maxRank ) then
				color = INVALID;
			elseif ( rank < goalRank ) then
				if ( freePoints > 0 and isGlowing ) then
					color = BUYMENOW;
				else
					if ( goalRank == maxRank ) then
						color = MAXPLANNED;
					else
						color = PLANNED;
					end
				end
			else
				if ( rank == maxRank ) then
					color = MAXEDOUT;
				elseif ( freePoints > 0 and isGlowing ) then
					color = DONTBUYME;
				else
					if ( goalRank == 0 ) then
						if ( Planner_CanInc( page, id ) ) then
							color = NORMAL;
						end
					else
						color = MAXEDOUT;
					end
				end
			end

			if ( color ) then
				getglobal( "PlayerTalentFrameTalent"..id.."Slot" ):SetVertexColor( color.r, color.g, color.b );
				getglobal( "PlayerTalentFrameTalent"..id.."RankBorder" ):Show();
				getglobal( "PlayerTalentFrameTalent"..id.."Rank" ):SetTextColor( color.r, color.g, color.b );
				getglobal( "PlayerTalentFrameTalent"..id.."Rank" ):SetText( rank.."/"..maxRank.." " );
				getglobal( "PlayerTalentFrameTalent"..id.."Rank" ):Show();
			else
				getglobal( "PlayerTalentFrameTalent"..id.."RankBorder" ):Hide();
				getglobal( "PlayerTalentFrameTalent"..id.."Rank" ):Hide();
			end
			
			if ( color and goalRank > 0 ) then
				getglobal( "PlannerRankBorder"..id ):Show();
				getglobal( "PlannerRank"..id ):SetTextColor( color.r, color.g, color.b );
				getglobal( "PlannerRank"..id ):SetText( goalRank );
				getglobal( "PlannerRank"..id ):Show();
			else
				getglobal( "PlannerRankBorder"..id ):Hide();
				getglobal( "PlannerRank"..id ):Hide();
			end

			local pagePointsText = HIGHLIGHT_FONT_COLOR_CODE..points.." ("..pagePlanned.." "..PLANNER_PLANNED..")"..FONT_COLOR_CODE_CLOSE;
			PlayerTalentFrameSpentPointsText:SetText( "This tab: "..pagePointsText );

			local pointsToSpend = freePoints.." ("..(PLANNER_MAX_POINTS - totalPlanned).." "..PLANNER_UNPLANNED..")";
			PlayerTalentFrameTalentPointsText:SetText( "Unspent: "..pointsToSpend );
		else
			getglobal( "PlannerRankBorder"..id ):Hide();
			getglobal( "PlannerRank"..id ):Hide();
		end
	end
end

function Planner_OnClick( self, button )
	local page = PanelTemplates_GetSelectedTab( PlayerTalentFrame );
	local id = self:GetID();
	local inspect = PlayerTalentFrame.inspect;
	local pet = PlayerTalentFrame.pet;
	local spec = PlayerTalentFrame.talentGroup;
	
	if ( tp_standby or not tp_name or not tp_class or inspect or pet ) then
		return Planner_TalentFrameTalent_OnClick( self, button );
	end

	if ( IsModifiedClick( "CHATLINK" ) or IsControlKeyDown() ) then
		Planner_TalentFrameTalent_OnClick( self, button );
		Planner_OnEnter( self );
		return;
	end
	
	if ( button == "LeftButton" ) then
		Planner_IncTalent( self, page, id );
	elseif ( button == "RightButton" ) then
		Planner_DecTalent( self, page, id );
	end
end

function Planner_OnEnter( self )
	Planner_TalentFrameTalent_OnEnter( self );

	local page = PanelTemplates_GetSelectedTab( PlayerTalentFrame );
	local id = self:GetID();
	local inspect = PlayerTalentFrame.inspect;
	local pet = PlayerTalentFrame.pet;
	local spec = PlayerTalentFrame.talentGroup;
	
	if ( tp_standby or not tp_name or not tp_class or inspect or pet ) then
		return;
	end

	local _, _, _, _, actRank, maxRank, _, _, pvwRank = GetTalentInfo( page, id, nil, nil, spec );
	local rank = GetCVarBool("previewTalents") and pvwRank or actRank;
	local talent = tp_webdata[tp_class][page][id];
	local goalRank = Planner_GetRank( page, id );
	
	local text = "";
	
	-- Hide the existing click messages if we're on the active spec
	if ( spec == GetActiveTalentGroup() ) then
		for i = 1, GameTooltip:NumLines(), 1 do
			text = getglobal( "GameTooltipTextLeft"..i ):GetText();
			
			if ( text == "Click to learn" or text == "Left click to add a point" or text == "Right click to remove a point" ) then
				getglobal( "GameTooltipTextLeft"..i ):SetText(nil);
			end
		end
	end
	
	if ( ( goalRank == 1 and rank == 0 ) or ( goalRank == maxRank and rank == maxRank ) ) then
		local line2 = getglobal( "GameTooltipTextLeft2" );
		line2:SetText( PLANNER_PLANNED_RANK..goalRank.."/"..maxRank );
	elseif ( ( goalRank > 0 and goalRank < maxRank ) or ( goalRank == maxRank and rank < maxRank ) ) then
		GameTooltip:AddLine( " " );
		GameTooltip:AddLine( PLANNER_PLANNED_RANK..goalRank.."/"..maxRank, 1.0, 1.0, 1.0 );
		GameTooltip:AddLine( talent[goalRank], 1.0, 0.82, 0.0, 1.0, 1 );
	end
	
	GameTooltip:Show();
end

function Planner_IncTalent( self, page, id )
	local goalRank = Planner_GetRank( page, id );
	local spec = PlayerTalentFrame.talentGroup;

	if ( Planner_CanInc( page, id ) ) then
		tp_talents[tp_name][spec][page][id] = goalRank + 1;
		TalentFrame_Update( PlayerTalentFrame );
		Planner_OnEnter( self );
	end
end

function Planner_DecTalent( self, page, id )
	local goalRank = Planner_GetRank( page, id );
	local spec = PlayerTalentFrame.talentGroup;
	
	if ( Planner_CanDec( page, id ) ) then
		if ( goalRank > 1 ) then
			tp_talents[tp_name][spec][page][id] = goalRank - 1;
		else
			tp_talents[tp_name][spec][page][id] = nil;
		end

		TalentFrame_Update( PlayerTalentFrame );
		Planner_OnEnter( self );
	end
end

function Planner_CanInc( page, id )
	local _, _, row, _, _, maxRank = GetTalentInfo( page, id );
	
	local talent = tp_webdata[tp_class][page][id];
	local goalRank = Planner_GetRank( page, id );
	local pagePlanned = Planner_GetCount( page, row );
	local totalPlanned = Planner_GetCount();
	local spec = PlayerTalentFrame.talentGroup;
	
	-- Ensure we've not already maxed this talent and that we have points
	if ( goalRank >= maxRank or totalPlanned >= PLANNER_MAX_POINTS ) then
		return nil;
	end
	
	-- Ensure we've spent enough points to be on this row
	if ( ( (row - 1) * 5 > pagePlanned ) ) then
		return nil;
	end
	
	-- Ensure we've done the listed prereqs
	local reqIndex, _, reqNum = Planner_GetPrereq( page, id );
	
	if ( reqIndex and tp_talents[tp_name][spec][page][reqIndex] ~= reqNum ) then
		return nil;
	end
	
	return true;
end

function Planner_CanDec( page, id )
	local _, _, row, col = GetTalentInfo( page, id );
	
	local talent = tp_webdata[tp_class][page][id];
	local goalRank = Planner_GetRank( page, id );
	local spec = PlayerTalentFrame.talentGroup;
	
	-- Ensure we have goal ranks in this talent
	if ( goalRank <= 0 ) then
		return nil;
	end
	
	-- Ensure we don't need this talent for future rows
	local numTalents = GetNumTalents( page );
	
	for i = id, numTalents, 1 do
		local reqName, _, tier = GetTalentInfo( page, i );
		local points = Planner_GetCount( page, tier );
		
		if ( tp_talents[tp_name][spec][page][i] and row < tier and points <= (tier - 1) * 5 ) then
			return nil;
		end
		
		local reqRow, reqCol = GetTalentPrereqs( page, i );
		
		if ( reqRow and reqCol ) then
			if ( row == reqRow and col == reqCol and tp_talents[tp_name][spec][page][i] ) then
				return nil;
			end
		end
	end
	
	return true;
end

function Planner_Upgrade()
	if ( tp_ver ~= 3.6 ) then
		tp_talents = {};
		tp_saved = {};
		tp_ver = 3.6;
	end
end

function Planner_Log( text )
	SELECTED_CHAT_FRAME:AddMessage( text );
end

function Planner_Debug( text, r, g, b )
	ChatFrame3:AddMessage( text, r, g, b );
end
